--- ./configure.ac.orig	2010-06-19 20:25:14.000000000 +0300
+++ ./configure.ac	2010-06-25 19:24:13.000000000 +0300
@@ -256,6 +256,26 @@
 fi
 
 
+# libSM for X session management
+SM_LIBS=""
+AC_ARG_ENABLE(libsm,
+	[  --enable-libsm          enable X session management support [[]]],
+	[enable_libsm=$enableval], [enable_libsm=yes])
+AC_MSG_CHECKING([whether to use LibSM])
+if test x"$enable_libsm" = xyes; then
+	AC_MSG_RESULT(yes)
+	AC_CHECK_HEADERS(X11/SM/SMlib.h, [SM_LIBS="-lSM -lICE"], enable_libsm=no)
+	if test x"$enable_libsm" = xyes; then
+		AC_DEFINE(HAVE_LIBSM, 1, [Define to 1 if you have libSM installed])
+	else
+		AC_MSG_WARN([X session management will not be supported])
+	fi
+else
+	AC_MSG_RESULT(no)
+fi
+AC_SUBST(SM_LIBS)
+
+
 # Set ${datadir}
 if test "x${datadir}" = 'x${prefix}/share' -o "x${datarootdir}" = 'x${prefix}/share'; then
 	if test "x${prefix}" = "xNONE"; then
--- ./doc/Doxyfile.in.orig	2010-06-19 20:19:29.000000000 +0300
+++ ./doc/Doxyfile.in	2010-06-25 19:24:13.000000000 +0300
@@ -229,7 +229,7 @@
 INCLUDE_PATH           =
 INCLUDE_FILE_PATTERNS  =
 # make G_GNUC_PRINTF a no-op unless doxygen would ignore functions with varargs
-PREDEFINED             = "G_GNUC_PRINTF(x,y)=" HAVE_PLUGINS GEANY_FUNCTIONS_H
+PREDEFINED             = "G_GNUC_PRINTF(x,y)=" HAVE_PLUGINS GEANY_FUNCTIONS_H HAVE_LIBSM
 EXPAND_AS_DEFINED      =
 SKIP_FUNCTION_MACROS   = NO
 #---------------------------------------------------------------------------
--- ./src/Makefile.am.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/Makefile.am	2010-06-25 19:24:13.000000000 +0300
@@ -38,6 +38,7 @@
 	project.c project.h \
 	sciwrappers.c sciwrappers.h \
 	search.c search.h \
+	sm.c sm.h \
 	socket.c socket.h \
 	stash.c stash.h \
 	support.h \
@@ -113,7 +114,7 @@
 
 geany_SOURCES = $(SRCS) vte.c vte.h
 
-geany_LDADD = ../scintilla/libscintilla.a ../tagmanager/libtagmanager.a @GTK_LIBS@ @GIO_LIBS@ $(INTLLIBS)
+geany_LDADD = ../scintilla/libscintilla.a ../tagmanager/libtagmanager.a @GTK_LIBS@ @GIO_LIBS@ @SM_LIBS@ $(INTLLIBS)
 
 AM_CFLAGS = -DGEANY_DATADIR=\""$(datadir)"\" \
 			-DGEANY_DOCDIR=\""$(docdir)"\" \
--- ./src/callbacks.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/callbacks.c	2010-06-25 19:24:13.000000000 +0300
@@ -87,21 +87,6 @@
 /*static gboolean switch_tv_notebook_page = FALSE; */
 
 
-static gboolean check_no_unsaved(void)
-{
-	guint i;
-
-	for (i = 0; i < documents_array->len; i++)
-	{
-		if (documents[i]->is_valid && documents[i]->changed)
-		{
-			return FALSE;
-		}
-	}
-	return TRUE;	/* no unsaved edits */
-}
-
-
 /* set editor_info.click_pos to the current cursor position if insert_callback_from_menu is TRUE
  * to prevent invalid cursor positions which can cause segfaults */
 static void verify_click_pos(GeanyDocument *doc)
@@ -117,7 +102,7 @@
 /* should only be called from on_exit_clicked */
 static void quit_app(void)
 {
-	configuration_save();
+	configuration_save(NULL);
 
 	if (app->project != NULL)
 		project_close(FALSE);	/* save project session files */
@@ -134,9 +119,12 @@
 gboolean
 on_exit_clicked                        (GtkWidget *widget, gpointer gdata)
 {
+	if (main_status.prevent_interaction)
+		return FALSE;
+
 	main_status.quitting = TRUE;
 
-	if (! check_no_unsaved())
+	if (document_any_unsaved())
 	{
 		if (document_account_for_unsaved())
 		{
--- ./src/document.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/document.c	2010-06-25 19:24:13.000000000 +0300
@@ -2822,12 +2822,23 @@
 }
 
 
-/* @note If successful, this should always be followed up with a call to
- * document_close_all().
- * @return TRUE if all files were saved or had their changes discarded. */
-gboolean document_account_for_unsaved(void)
+/* @return TRUE if there are some unsaved documents */
+gboolean document_any_unsaved(void)
+{
+	guint i;
+
+	foreach_document(i)
+		if (documents[i]->changed)
+			return TRUE;
+	return FALSE;
+}
+
+
+/* @return TRUE if all files were saved or had their changes discarded.
+ * The files with discarded changes remain marked as modified. */
+gboolean document_prompt_for_unsaved(void)
 {
-	guint i, p, page_count, len = documents_array->len;
+	guint p, page_count;
 	GeanyDocument *doc;
 
 	page_count = gtk_notebook_get_n_pages(GTK_NOTEBOOK(main_widgets.notebook));
@@ -2840,6 +2851,20 @@
 				return FALSE;
 		}
 	}
+	return TRUE;
+}
+
+
+/* @note If successful, this should always be followed up with a call to
+ * document_close_all().
+ * @return TRUE if all files were saved or had their changes discarded. */
+gboolean document_account_for_unsaved(void)
+{
+	guint i, len = documents_array->len;
+	GeanyDocument *doc;
+
+	if (!document_prompt_for_unsaved())
+		return FALSE;
 	/* all documents should now be accounted for, so ignore any changes */
 	for (i = 0; i < len; i++)
 	{
--- ./src/document.h.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/document.h	2010-06-25 19:24:13.000000000 +0300
@@ -197,6 +197,10 @@
 
 gboolean document_close(GeanyDocument *doc);
 
+gboolean document_any_unsaved(void);
+
+gboolean document_prompt_for_unsaved(void);
+
 gboolean document_account_for_unsaved(void);
 
 gboolean document_close_all(void);
--- ./src/keyfile.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/keyfile.c	2010-06-25 19:24:13.000000000 +0300
@@ -533,10 +533,10 @@
 }
 
 
-void configuration_save(void)
+void configuration_save(const gchar *libsm_client_id)
 {
 	GKeyFile *config = g_key_file_new();
-	gchar *configfile = g_strconcat(app->configdir, G_DIR_SEPARATOR_S, "geany.conf", NULL);
+	gchar *configfile = configuration_name(libsm_client_id);
 	gchar *data;
 
 	g_key_file_load_from_file(config, configfile, G_KEY_FILE_NONE, NULL);
@@ -550,7 +550,7 @@
 	save_recent_files(config, ui_prefs.recent_queue, "recent_files");
 	save_recent_files(config, ui_prefs.recent_projects_queue, "recent_projects");
 
-	if (cl_options.load_session)
+	if (cl_options.load_session || libsm_client_id)
 		configuration_save_session_files(config);
 
 	/* write the file */
@@ -894,7 +894,7 @@
  */
 void configuration_save_default_session(void)
 {
-	gchar *configfile = g_strconcat(app->configdir, G_DIR_SEPARATOR_S, "geany.conf", NULL);
+	gchar *configfile = configuration_name(NULL);
 	gchar *data;
 	GKeyFile *config = g_key_file_new();
 
@@ -918,7 +918,7 @@
  */
 void configuration_reload_default_session(void)
 {
-	gchar *configfile = g_strconcat(app->configdir, G_DIR_SEPARATOR_S, "geany.conf", NULL);
+	gchar *configfile = configuration_name(NULL);
 	GKeyFile *config = g_key_file_new();
 
 	g_key_file_load_from_file(config, configfile, G_KEY_FILE_NONE, NULL);
@@ -930,11 +930,17 @@
 }
 
 
-gboolean configuration_load(void)
+gboolean configuration_load(const gchar *libsm_client_id)
 {
-	gchar *configfile = g_strconcat(app->configdir, G_DIR_SEPARATOR_S, "geany.conf", NULL);
+	gchar *configfile = configuration_name(libsm_client_id);
 	GKeyFile *config = g_key_file_new();
 
+	if (libsm_client_id && ! g_file_test(configfile, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_IS_SYMLINK))
+	{	/* session config file does not exist, so try falling back to geany.conf */
+		geany_debug("No session config file found, try to use default configuration.");
+		setptr(configfile, configuration_name(NULL));
+	}
+
 	if (! g_file_test(configfile, G_FILE_TEST_IS_REGULAR | G_FILE_TEST_IS_SYMLINK))
 	{	/* config file does not (yet) exist, so try to load a global config file which may be */
 		/* created by distributors */
@@ -1122,3 +1128,16 @@
 	g_ptr_array_free(keyfile_groups, TRUE);
 	g_ptr_array_free(pref_groups, TRUE);
 }
+
+
+gchar *configuration_name(const gchar *suffix)
+{
+	gchar *configfile;
+
+	if (suffix)
+		configfile = g_strconcat(app->configdir, G_DIR_SEPARATOR_S, "geany-", suffix, ".conf", NULL);
+	else
+		configfile = g_strconcat(app->configdir, G_DIR_SEPARATOR_S, "geany.conf", NULL);
+
+	return configfile;
+}
--- ./src/keyfile.h.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/keyfile.h	2010-06-25 19:24:13.000000000 +0300
@@ -37,9 +37,12 @@
 
 void configuration_add_pref_group(struct StashGroup *group, gboolean for_prefs_dialog);
 
-void configuration_save(void);
+/* return newly allocated "app->configdir/geany[-suffix].conf" name */
+gchar *configuration_name(const gchar *suffix);
 
-gboolean configuration_load(void);
+void configuration_save(const gchar *libsm_client_id);
+
+gboolean configuration_load(const gchar *libsm_client_id);
 
 void configuration_open_files(void);
 
--- ./src/main.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/main.c	2010-06-25 19:24:13.000000000 +0300
@@ -74,6 +74,7 @@
 #include "printing.h"
 #include "toolbar.h"
 #include "geanyobject.h"
+#include "sm.h"
 
 #ifdef HAVE_SOCKET
 # include "socket.h"
@@ -115,9 +116,17 @@
 static gboolean no_plugins = FALSE;
 #endif
 static gboolean dummy = FALSE;
+static gchar * libsm_client_id = NULL;
 
+/* WARNING: Do not change values of variables where values of command-line options are stored!
+ * This is, they should remain unchanged after `g_option_context_parse' returns.
+ * Otherwise, "restart command" for X session management may be filled improperly.
+ *
+ * NOTE: Currently optentries of type G_OPTION_ARG_CALLBACK are not supported by
+ * X session management support implementation.
+ */
 /* in alphabetical order of short options */
-static GOptionEntry entries[] =
+GOptionEntry optentries[] =
 {
 	{ "column", 0, 0, G_OPTION_ARG_INT, &cl_options.goto_column, N_("Set initial column number for the first opened file (useful in conjunction with --line)"), NULL },
 	{ "config", 'c', 0, G_OPTION_ARG_FILENAME, &alternate_config, N_("Use an alternate configuration directory"), NULL },
@@ -144,9 +153,40 @@
 	{ "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose_mode, N_("Be verbose"), NULL },
 	{ "version", 'V', 0, G_OPTION_ARG_NONE, &show_version, N_("Show version and exit"), NULL },
 	{ "dummy", 0, G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_NONE, &dummy, NULL, NULL }, /* for +NNN line number arguments */
+	{ "libsm-client-id", 0, G_OPTION_FLAG_HIDDEN, G_OPTION_ARG_STRING, &libsm_client_id, NULL, NULL },
+	/* add new options here and in `optentries_aux' below */
 	{ NULL, 0, 0, 0, NULL, NULL, NULL }
 };
 
+GeanyOptionEntryAux optentries_aux[] = {
+	{FALSE}, /* "column" */
+	{TRUE},  /* "config" */
+	{FALSE}, /* "ft-names */
+	{FALSE}, /* "generate-tags" */
+	{FALSE}, /* "no-preprocessing" */
+#ifdef HAVE_SOCKET
+	{TRUE},  /* "new-instance" */
+	{TRUE},  /* "socket-file" */
+	{FALSE}, /* "list-documents" */
+#endif
+	{FALSE}, /* "line" */
+	{TRUE},  /* "no-msgwin" */
+	{TRUE},  /* "no-ctags" */
+#ifdef HAVE_PLUGINS
+	{TRUE},  /* "no-plugins" */
+#endif
+	{FALSE}, /* "print-prefix" */
+	{FALSE}, /* "no-session" */
+#ifdef HAVE_VTE
+	{TRUE},  /* "no-terminal" */
+	{TRUE},  /* "vte-lib" */
+#endif
+	{TRUE},  /* "verbose" */
+	{FALSE}, /* "version" */
+	{FALSE}, /* "dummy" */
+	{FALSE}, /* "libsm-client-id" option is handled separately */
+};
+
 
 static void setup_window_position(void)
 {
@@ -496,7 +536,7 @@
 	}
 
 	context = g_option_context_new(_("[FILES...]"));
-	g_option_context_add_main_entries(context, entries, GETTEXT_PACKAGE);
+	g_option_context_add_main_entries(context, optentries, GETTEXT_PACKAGE);
 	g_option_group_set_translation_domain(g_option_context_get_main_group(context), GETTEXT_PACKAGE);
 	g_option_context_add_group(context, gtk_get_option_group(FALSE));
 	g_option_context_parse(context, argc, argv, &error);
@@ -538,7 +578,7 @@
 	if (alternate_config)
 	{
 		geany_debug("alternate config: %s", alternate_config);
-		app->configdir = alternate_config;
+		app->configdir = g_strdup(alternate_config);
 	}
 	else
 	{
@@ -820,7 +860,7 @@
 	locale_filename = utils_get_locale_from_utf8(project_prefs.session_file);
 
 	if (NZV(locale_filename))
-		project_load_file(locale_filename);
+		project_load_file(locale_filename, libsm_client_id == NULL);
 
 	g_free(locale_filename);
 	g_free(project_prefs.session_file);	/* no longer needed */
@@ -829,7 +869,7 @@
 
 static void load_settings(void)
 {
-	configuration_load();
+	configuration_load(libsm_client_id);
 	/* let cmdline options overwrite configuration settings */
 #ifdef HAVE_VTE
 	vte_info.have_vte = (no_vte) ? FALSE : vte_info.load_vte;
@@ -857,7 +897,7 @@
 		if (use_session)
 			project_load_file_with_session(pfile);
 		else
-			project_load_file(pfile);
+			project_load_file(pfile, TRUE);
 	}
 	g_free(pfile);
 }
@@ -874,18 +914,20 @@
 
 	if (load_project_from_cl || ! open_cl_files(argc, argv))
 	{
-		if (prefs.load_session)
+		if (prefs.load_session || libsm_client_id)
 		{
+			gboolean new_session = cl_options.new_instance && !libsm_client_id;
+
 			if (load_project_from_cl)
 			{
 				main_load_project_from_command_line(argv[1], FALSE);
 			}
-			else if (cl_options.load_session && !cl_options.new_instance)
+			else if (cl_options.load_session && !new_session)
 				load_session_project_file();
 
 			/* when we want a new instance, we still load project session files unless -s
 			 * was passed */
-			if (!cl_options.load_session || (!load_project_from_cl && cl_options.new_instance))
+			if (!cl_options.load_session || (!load_project_from_cl && new_session))
 				return;
 
 			/* load session files into tabs, as they are found in the session_files variable */
@@ -992,6 +1034,7 @@
 	geany_object = geany_object_new();
 
 	/* inits */
+	sm_init(argv[0], libsm_client_id);
 	main_init();
 
 	encodings_init();
@@ -1135,6 +1178,8 @@
 {
 	geany_debug("Quitting...");
 
+	sm_finalize();
+
 #ifdef HAVE_SOCKET
 	socket_finalize();
 #endif
--- ./src/main.h.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/main.h	2010-06-25 19:24:13.000000000 +0300
@@ -25,6 +25,20 @@
 #ifndef GEANY_MAIN_H
 #define GEANY_MAIN_H
 
+/** Information about command line entries that can not be stored in GOptionEntry. */
+typedef struct
+{
+	/**
+	 * Indicates whether the value of a command-line option should go to "restart command".
+	 * See X session management support implementation in sm.{c,h}.
+	 */
+	gboolean persist_upon_restart;
+}
+GeanyOptionEntryAux;
+
+extern GOptionEntry optentries[];
+extern GeanyOptionEntryAux optentries_aux[];
+
 typedef struct
 {
 	gboolean	new_instance;
@@ -47,6 +61,7 @@
 							  * (used to prevent notebook switch page signals) */
 	gboolean	quitting;	/* state when Geany is quitting completely */
 	gboolean	main_window_realized;
+	gboolean	prevent_interaction;	/* state while saving SM session; the main window is locked */
 }
 GeanyStatus;
 
--- ./src/makefile.win32.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/makefile.win32	2010-06-25 19:24:13.000000000 +0300
@@ -66,7 +66,7 @@
 		interface.o keybindings.o keyfile.o \
 		log.o main.o msgwindow.o navqueue.o notebook.o plugins.o pluginutils.o \
 		prefs.o printing.o project.o \
-		sciwrappers.o search.o socket.o stash.o \
+		sciwrappers.o search.o sm.o socket.o stash.o \
 		symbols.o templates.o toolbar.o tools.o sidebar.o \
 		ui_utils.o utils.o win32.o
 
--- ./src/prefs.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/prefs.c	2010-06-25 19:24:13.000000000 +0300
@@ -1175,7 +1175,7 @@
 		ui_update_view_editor_menu_items();
 
 		/* store all settings */
-		configuration_save();
+		configuration_save(NULL);
 	}
 
 	if (response != GTK_RESPONSE_APPLY)
--- ./src/project.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/project.c	2010-06-25 19:24:13.000000000 +0300
@@ -82,7 +82,7 @@
 
 static gboolean update_config(const PropertyDialogElements *e);
 static void on_file_save_button_clicked(GtkButton *button, PropertyDialogElements *e);
-static gboolean load_config(const gchar *filename);
+static gboolean load_config(const gchar *filename, gboolean load_session);
 static gboolean write_config(gboolean emit_signal);
 static void on_name_entry_changed(GtkEditable *editable, PropertyDialogElements *e);
 static void on_entries_changed(GtkEditable *editable, PropertyDialogElements *e);
@@ -202,7 +202,7 @@
 
 gboolean project_load_file_with_session(const gchar *locale_file_name)
 {
-	if (project_load_file(locale_file_name))
+	if (project_load_file(locale_file_name, TRUE))
 	{
 		if (project_prefs.project_session)
 		{
@@ -936,11 +936,11 @@
 }
 
 
-gboolean project_load_file(const gchar *locale_file_name)
+gboolean project_load_file(const gchar *locale_file_name, gboolean load_session)
 {
 	g_return_val_if_fail(locale_file_name != NULL, FALSE);
 
-	if (load_config(locale_file_name))
+	if (load_config(locale_file_name, load_session))
 	{
 		gchar *utf8_filename = utils_get_utf8_from_locale(locale_file_name);
 
@@ -964,8 +964,9 @@
 /* Reads the given filename and creates a new project with the data found in the file.
  * At this point there should not be an already opened project in Geany otherwise it will just
  * return.
- * The filename is expected in the locale encoding. */
-static gboolean load_config(const gchar *filename)
+ * The filename is expected in the locale encoding. Both project_session preference and
+ * load_session must be TRUE for the project session files to be loaded. */
+static gboolean load_config(const gchar *filename, gboolean load_session)
 {
 	GKeyFile *config;
 	GeanyProject *p;
@@ -997,7 +998,7 @@
 	apply_editor_prefs();
 
 	build_load_menu(config, GEANY_BCS_PROJ, (gpointer)p);
-	if (project_prefs.project_session)
+	if (project_prefs.project_session && load_session)
 	{
 		/* save current (non-project) session (it could has been changed since program startup) */
 		configuration_save_default_session();
--- ./src/project.h.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/project.h	2010-06-25 19:24:13.000000000 +0300
@@ -77,7 +77,7 @@
 gboolean project_ask_close(void);
 
 
-gboolean project_load_file(const gchar *locale_file_name);
+gboolean project_load_file(const gchar *locale_file_name, gboolean load_session);
 
 gboolean project_load_file_with_session(const gchar *locale_file_name);
 
--- ./src/sm.c.orig	2010-06-25 19:24:13.000000000 +0300
+++ ./src/sm.c	2010-06-25 19:24:13.000000000 +0300
@@ -0,0 +1,522 @@
+/*
+ *      sm.c - this file is part of Geany, a fast and lightweight IDE
+ *
+ *      Copyright 2009 Eugene Arshinov <earshinov(at)gmail(dot)com>
+ *
+ *      This program is free software; you can redistribute it and/or modify
+ *      it under the terms of the GNU General Public License as published by
+ *      the Free Software Foundation; either version 2 of the License, or
+ *      (at your option) any later version.
+ *
+ *      This program is distributed in the hope that it will be useful,
+ *      but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *      GNU General Public License for more details.
+ *
+ *      You should have received a copy of the GNU General Public License
+ *      along with this program; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * $Id$
+ */
+
+/*
+ * @file sm.c
+ * Provides X session management protocol (XSMP) support using libSM library.
+ *
+ * In order to support XSMP, we have to support Inter-Client Exchange
+ * Protocol (ICE). This file takes care of the latter too.
+ *
+ * Typical usage: @c sm_init() is called when Geany is starting and
+ * @c sm_finalize() is called when Geany is quitting.
+ *
+ * According to libSM documentation, client should retain the same ID after
+ * it is restarted. The main module (@c main.c) maintains "--libsm-client-id"
+ * command-line option and passes the specified value (if any) to @c sm_init().
+ */
+
+/*
+ * As libSM is not available on Windows,
+ * it is safe enough to use POSIX-specific things here.
+ */
+
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#endif
+
+#ifdef HAVE_LIBSM
+
+#include <stdlib.h>
+#include <string.h>
+
+#include <glib.h>
+#include <gtk/gtk.h>
+
+#include <X11/SM/SMlib.h>
+#include <X11/ICE/ICElib.h>
+
+#include "geany.h"
+#include "main.h" /* for cl_options */
+#include "ui_utils.h"  /* for main_widgets */
+#include "document.h"
+#include "utils.h" /* for option reverse parsing */
+#include "keyfile.h"
+
+
+static void ice_connection_watch(IceConn icecon, IcePointer client_data, Bool opening,
+	IcePointer *watch_data);
+static gboolean ice_iochannel_watch(GIOChannel * source, GIOCondition condition, gpointer data);
+
+static SmcConn sm_connect(const char * libsm_client_id);
+static void sm_set_props(const char * argv0);
+static void sm_prevent_interaction(gboolean prevent);
+
+static void sm_save_yourself_callback(SmcConn smcon, SmPointer client_data,
+	int save_type, Bool shutdown, int interact_style, Bool fast);
+static void sm_interact_callback(SmcConn smcon, SmPointer client_data);
+static void sm_save_complete_callback(SmcConn smcon, SmPointer client_data);
+static void sm_shutdown_cancelled_callback(SmcConn smcon, SmPointer client_data);
+static void sm_die_callback(SmcConn smcon, SmPointer client_data);
+
+
+/* LibSM connection object initialized in @c sm_init() and used in @c sm_finalize(). */
+static SmcConn smc_conn;
+
+/* LibSM new client id. Initialized by @c sm_init(), used in many places. */
+static char *new_client_id = NULL;
+
+/* The last save_type and shutdown values received by sm_save_yourself_callback().
+ * Used by the sm_interact_callback(). */
+static int sm_last_save_type;
+static Bool sm_last_shutdown;
+
+#endif
+
+/*
+ * @name Exported functions
+ * @{
+ */
+
+/*
+ * Initialize XSMP support.
+ *
+ * @param argv0             Value of @c argv[0] used to define Geany's restart command.
+ * @param libsm_client_id   Client-ID specified with "--libsm-client-id" command line
+ *                          option or @c NULL if the option was not passed.
+ *
+ * This function connects to the session manager using @c sm_connect(). If
+ * everything is successful, it stores libSM connection object and the new session id
+ * in the global variables @c smcon and @c new_client_id, and calls @c sm_set_props().
+ *
+ * This function can be called at any time during Geany instance's lifetime.
+ * Actually it is called when Geany is starting.
+ *
+ * When Geany is compiled without XSMP support, this function is a no-op.
+ */
+void sm_init(const char * argv0, const char * libsm_client_id)
+{
+	#ifdef HAVE_LIBSM
+	/* This function should be called once */
+	g_assert(!smc_conn);
+	if (smc_conn)
+		return;
+
+	smc_conn = sm_connect(libsm_client_id);
+	if (!smc_conn)
+		return;
+
+	gdk_set_sm_client_id(new_client_id);
+	sm_set_props(argv0);
+	#endif
+}
+
+
+/*
+ * Perform cleanup.
+ *
+ * Call this function when XSMP support is no longer needed. In fact it is
+ * called when Geany is quitting.
+ *
+ * When Geany is compiled without XSMP support, this function is a no-op.
+ */
+void sm_finalize(void)
+{
+	#ifdef HAVE_LIBSM
+	if (smc_conn)
+	{
+		free(new_client_id);
+		SmcCloseConnection(smc_conn, 0, NULL);
+		smc_conn = 0;
+	}
+	#endif
+}
+
+
+/* @} */
+
+#ifdef HAVE_LIBSM
+
+/*
+ * @name ICE support
+ * @{
+ */
+
+/*
+ * ICE connection watcher used to attach a GIOChannel to each ICE connection
+ * so that this connection can be handled in GTK main loop.
+ *
+ * @param icecon        ICE connection.
+ * @param client_data   Client data specified in @c IceAddConnectionWatch() function call.
+ *                      Currently it is not used.
+ * @param opening       Whether @c icecon is opening or closing.
+ * @param watch_data    A piece of data that can be set when @c icecon is opening and
+ *                      read when it is closing. We store GIOChannel watcher ID here.
+ *
+ * We attach @c ice_iochannel_watch GIOChannel watcher to every created GIOChannel
+ * in order to handle messages. This is how session manager communicates to Geany.
+ *
+ * @see sm_connect()
+ */
+static void ice_connection_watch(IceConn icecon, IcePointer client_data,
+		Bool opening, IcePointer *watch_data)
+{
+	guint input_id;
+
+	if (opening)
+	{
+		GIOChannel * channel = g_io_channel_unix_new(IceConnectionNumber(icecon));
+		input_id = g_io_add_watch(channel, G_IO_IN | G_IO_HUP | G_IO_ERR,
+			ice_iochannel_watch, icecon);
+		g_io_channel_unref(channel);
+
+		*watch_data = (IcePointer)GUINT_TO_POINTER(input_id);
+	}
+	else
+	{
+		input_id = GPOINTER_TO_UINT((gpointer)*watch_data);
+		g_source_remove(input_id);
+	}
+}
+
+
+/*
+ * A watcher attached to a GIOChannel corresponding to an ICE connection.
+ *
+ * @param source   A GIOChannel corresponding to an ICE connection.
+ * @param condition
+ * @param data     Client data specified in @c g_io_add_watch() function call.
+ *                 An ICE connection object (of type @c IceConn) is stored here.
+ * @return         Return FALSE to remove the GIOChannel.
+ *
+ * This function calls @c IceProcessMessages causing an appropriate libSM
+ * callback to be invoked.
+ *
+ * @see ice_connection_watch()
+ */
+static gboolean ice_iochannel_watch(GIOChannel * source, GIOCondition condition, gpointer data)
+{
+	IceConn icecon = (IceConn)data;
+	IceProcessMessages(icecon, NULL, NULL);
+	return TRUE;
+}
+
+
+/*
+ * @}
+ * @name libSM support implementation
+ * @{
+ */
+
+/*
+ * Connect to the session manager.
+ *
+ * @param libsm_client_id      LibSM client ID saved from the previous session
+ *                             or @c NULL if there was no previous session.
+ * @return libSM connection object or @c 0 if connection to the session manager fails.
+ */
+static SmcConn sm_connect(const char * libsm_client_id)
+{
+	static const SmcCallbacks callbacks = {
+		{sm_save_yourself_callback, NULL},
+		{sm_die_callback, NULL},
+		{sm_save_complete_callback, NULL},
+		{sm_shutdown_cancelled_callback, NULL}};
+
+	gchar err[256] = "";
+	SmcConn smcon;
+
+	if (!g_getenv("SESSION_MANAGER"))
+		return 0;
+
+	IceAddConnectionWatch(ice_connection_watch, NULL);
+
+	smcon = (gpointer)SmcOpenConnection(NULL, NULL,
+		SmProtoMajor, SmProtoMinor,
+		SmcSaveYourselfProcMask |
+		SmcDieProcMask |
+		SmcSaveCompleteProcMask |
+		SmcShutdownCancelledProcMask,
+		(SmcCallbacks *)&callbacks,
+		(char *)libsm_client_id, &new_client_id,
+		256, err);
+
+	if (!smcon)
+	{
+		g_warning("While connecting to session manager:\n%s.", err);
+
+		IceRemoveConnectionWatch(ice_connection_watch, NULL);
+		return 0;
+	}
+
+	return smcon;
+}
+
+
+/*
+ * Store global @c SmPropValue objects.
+ *
+ * @param argv0             Value of @c argv[0].
+ *
+ * This function sets all libSM properties.
+ */
+static void sm_set_props(const char * argv0)
+{
+	gchar * curdir = g_get_current_dir();
+	const gchar * username = g_get_user_name();
+	char * client_id_arg = g_strconcat("--libsm-client-id=", new_client_id, NULL);
+	gchar * configfile = configuration_name(new_client_id);
+
+	SmPropValue curdir_val = {
+		strlen(curdir),
+		curdir};
+
+	SmPropValue userid_val, program_val, client_id_val;
+	SmPropValue discard_val[3] = { { 2, "rm" }, { 2, "-f" },
+		{ strlen(configfile), configfile } };
+
+	SmProp program_prop = {SmProgram, SmARRAY8, 1, &program_val};
+	SmProp userid_prop = {SmUserID, SmARRAY8, 1, &userid_val};
+	SmProp curdir_prop = {SmCurrentDirectory, SmARRAY8, 1, &curdir_val};
+	SmProp restart_prop, clone_prop;
+	SmProp discard_prop = {SmDiscardCommand, SmLISTofARRAY8, 3, discard_val};
+
+	SmProp *props[4] = { &program_prop, &curdir_prop, &restart_prop, &discard_prop };
+
+	int i, arg_max, argc;
+	SmPropValue *arg_val;
+
+	/* -- end of declarations -- */
+
+	program_val.length = strlen(argv0);
+	program_val.value = (char *)argv0;
+	client_id_val.length = strlen(client_id_arg);
+	client_id_val.value = (char *)client_id_arg;
+
+	for (i = 0, arg_max = 2; optentries[i].long_name; i++)
+		if (optentries_aux[i].persist_upon_restart)
+			arg_max++;
+
+	arg_val = g_new(SmPropValue, arg_max);
+	arg_val[0] = program_val;
+	arg_val[1] = client_id_val;
+
+	for (i = 0, argc = 2; optentries[i].long_name; i++)
+	{
+		if (optentries_aux[i].persist_upon_restart)
+		{
+			char *option = utils_option_entry_reverse_parse(&optentries[i]);
+
+			if (option)
+			{
+				arg_val[argc].length = strlen(option);
+				arg_val[argc].value = option;
+				argc++;
+			}
+		}
+	}
+
+	restart_prop.name = SmRestartCommand;
+	restart_prop.type = SmLISTofARRAY8;
+	restart_prop.num_vals = argc;
+	restart_prop.vals = arg_val;
+
+	SmcSetProperties(smc_conn, 4, props);
+
+	arg_val[1] = program_val;
+	clone_prop.name = SmCloneCommand;
+	clone_prop.type = SmLISTofARRAY8;
+	clone_prop.num_vals = argc - 1;
+	clone_prop.vals = arg_val + 1;
+	props[0] = &clone_prop;
+	SmcSetProperties(smc_conn, 1, props);
+
+	if (username)
+	{
+		userid_val.length = strlen(username);
+		userid_val.value = (char *)username;
+		props[0] = &userid_prop;
+		SmcSetProperties(smc_conn, 1, props);
+	}
+
+	g_free(curdir);
+	g_free(client_id_arg);
+	g_free(configfile);
+	for (i = 2; i < argc; i++)
+		g_free(arg_val[i].value);
+	g_free(arg_val);
+}
+
+
+/*
+ * Prevent/allow user interaction with Geany.
+ *
+ * @param prevent  TRUE to prevent interaction, FALSE to allow.
+ */
+static void sm_prevent_interaction(gboolean prevent)
+{
+	if (main_status.main_window_realized)
+	{
+		gtk_widget_set_sensitive(main_widgets.window, prevent == FALSE);
+		main_status.prevent_interaction = prevent;
+	}
+}
+
+/*
+ * @}
+ * @name libSM callbacks
+ * @{
+ */
+
+/*
+ * "Save Yourself" callback.
+ *
+ * @param smcon            LibSM connection object.
+ * @param client_data      Client data specified when the callback was registered.
+ *                         Currently it is not used.
+ * @param save_type        Specifies the type of information that should be saved.
+ * @param shutdown         Specifies if a shutdown is taking place.
+ * @param interact_style   The type of interaction allowed with the user.
+ * @param fast             If True, the client should save its state as quickly as possible.
+ *
+ * See libSM documentation for more details.
+ *
+ * Here we save Geany's configuration. If there are any unsaved documents and
+ * interaction with the user is allowed, we also make an "Interact Request" so
+ * that session manager sends us an "Interact" message. Otherwise, we are ready
+ * to handle "Save Complete", "Die" or "Shutdown Cancelled" message.
+ *
+ * @see sm_interact_callback()
+ * @see sm_shutdown_cancelled_callback()
+ * @see sm_die_callback()
+ */
+static void sm_save_yourself_callback(SmcConn smcon, SmPointer client_data,
+		int save_type, Bool shutdown, int interact_style, Bool fast)
+{
+	if (save_type == SmSaveGlobal || save_type == SmSaveBoth)
+	{
+		if (interact_style == SmInteractStyleAny && document_any_unsaved())
+		{
+			if (!SmcInteractRequest(smcon, SmDialogNormal, sm_interact_callback, NULL))
+				SmcSaveYourselfDone(smcon, FALSE);
+			sm_last_save_type = save_type;
+			sm_last_shutdown = shutdown;
+			return;
+		}
+	}
+
+	if (save_type == SmSaveLocal || save_type == SmSaveBoth)
+		configuration_save(new_client_id);
+
+	SmcSaveYourselfDone(smcon, TRUE);
+	sm_prevent_interaction(TRUE);
+}
+
+
+/*
+ * "Interact" callback.
+ *
+ * @param smcon         LibSM connection object.
+ * @param client_data   Client data specified when the callback was registered.
+ *                      Currently it is not used.
+ *
+ * See libSM documentation for more details.
+ *
+ * The session manager sends us an "Interact" message after we make an
+ * "Interact Request" in @c sm_save_yourself_callback(). Here we are allowed to
+ * interact with the user, so we ask her whether to save changed documents
+ * (the user also can cancel the shutdown). After that we are ready to handle
+ * "Save Complete", "Shutdown Cancelled" or "Die" message.
+ *
+ * @see sm_shutdown_cancelled_callback()
+ * @see sm_die_callback()
+ */
+static void sm_interact_callback(SmcConn smcon, SmPointer client_data)
+{
+	gboolean unsaved = !document_prompt_for_unsaved();
+
+	SmcInteractDone(smcon, sm_last_shutdown && unsaved);
+	if (sm_last_save_type == SmSaveBoth)
+		configuration_save(new_client_id);
+	SmcSaveYourselfDone(smcon, TRUE);
+	sm_prevent_interaction(TRUE);
+}
+
+
+/*
+ * "Save Complete" callback.
+ *
+ * @param smcon         LibSM connection object.
+ * @param client_data   Client data specified when the callback was registered.
+ *                      Currently it is not used.
+ *
+ * See libSM documentation for more details.
+ *
+ * Unfreeze and continue normally.
+ */
+static void sm_save_complete_callback(SmcConn smcon, SmPointer client_data)
+{
+	sm_prevent_interaction(FALSE);
+}
+
+
+/**
+ * "Shutdown Cancelled" callback.
+ *
+ * @param smcon         LibSM connection object.
+ * @param client_data   Client data specified when the callback was registered.
+ *                      Currently it is not used.
+ *
+ * See libSM documentation for more details.
+ *
+ * Signal save done and unfreeze.
+ */
+static void sm_shutdown_cancelled_callback(SmcConn smcon, SmPointer client_data)
+{
+	SmcSaveYourselfDone(smcon, TRUE);
+	sm_prevent_interaction(FALSE);
+}
+
+
+/*
+ * "Die" callback.
+ *
+ * @param smcon         LibSM connection object.
+ * @param client_data   Client data specified when the callback was registered.
+ *                      Currently it is not used.
+ *
+ * See libSM documentation for more details.
+ *
+ * The session manager asks us to quit Geany and we do it. When quitting, the
+ * main module (@c main.c) will call @c sm_finalize() where we will close the
+ * connection to the session manager.
+ */
+static void sm_die_callback(SmcConn smcon, SmPointer client_data)
+{
+	main_status.quitting = TRUE;
+	main_quit();
+}
+
+
+/* @} */
+
+#endif
--- ./src/sm.h.orig	2010-06-25 19:24:13.000000000 +0300
+++ ./src/sm.h	2010-06-25 19:24:13.000000000 +0300
@@ -0,0 +1,30 @@
+/*
+ *      sm.h - this file is part of Geany, a fast and lightweight IDE
+ *
+ *      Copyright 2009 Eugene Arshinov <earshinov(at)gmail(dot)com>
+ *
+ *      This program is free software; you can redistribute it and/or modify
+ *      it under the terms of the GNU General Public License as published by
+ *      the Free Software Foundation; either version 2 of the License, or
+ *      (at your option) any later version.
+ *
+ *      This program is distributed in the hope that it will be useful,
+ *      but WITHOUT ANY WARRANTY; without even the implied warranty of
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ *      GNU General Public License for more details.
+ *
+ *      You should have received a copy of the GNU General Public License
+ *      along with this program; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * $Id$
+ */
+
+#ifndef GEANY_SM_H
+#define GEANY_SM_H 1
+
+void sm_init(const char * argv0, const char * libsm_client_id);
+
+void sm_finalize(void);
+
+#endif
--- ./src/utils.c.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/utils.c	2010-06-25 19:24:13.000000000 +0300
@@ -2007,3 +2007,47 @@
 }
 
 
+/*
+ * "Reverse parse" @c GOptionEntry
+ *
+ * @param optentry    Object to parse.
+ * @return            Newly-allocated option string or NULL.
+ *
+ * Function takes the information about the option entry stored in @c optentry
+ * and the option's value by the address stored in @c optentry. It returns a newly
+ * allocated option string, or NULL if none required (or if the option type is not
+ * supported). Any relative file names are converted to absolute.
+ *
+ * The resulting string must be freed with @c g_free().
+ */
+gchar *utils_option_entry_reverse_parse(const GOptionEntry * optentry)
+{
+	gchar *s = NULL;
+
+	switch (optentry->arg)
+	{
+		case G_OPTION_ARG_NONE:
+			{
+				gboolean val = *(gboolean *)optentry->arg_data;
+				gboolean reverse = (optentry->flags & G_OPTION_FLAG_REVERSE);
+				if ((val && !reverse) || (!val && reverse)) /* logical XOR */
+					s = g_strdup_printf("--%s", optentry->long_name);
+			}
+			break;
+
+		case G_OPTION_ARG_INT:
+			if (*(gint *)optentry->arg_data)
+				s = g_strdup_printf("--%s=%d", optentry->long_name, *(gint *)optentry->arg_data);
+			break;
+
+		case G_OPTION_ARG_STRING:
+		case G_OPTION_ARG_FILENAME:
+			if (*(gchar **)optentry->arg_data)
+				s = g_strdup_printf("--%s=%s", optentry->long_name, *(gchar **)optentry->arg_data);
+			break;
+
+		default:
+			g_warning("%s: %s: %d\n", G_STRFUNC, "Unsupported option entry type", optentry->arg);
+	}
+	return s;
+}
--- ./src/utils.h.orig	2010-06-19 20:18:49.000000000 +0300
+++ ./src/utils.h	2010-06-25 19:24:13.000000000 +0300
@@ -231,4 +231,6 @@
 
 gchar **utils_copy_environment(const gchar **exclude_vars, const gchar *first_varname, ...) G_GNUC_NULL_TERMINATED;
 
+gchar *utils_option_entry_reverse_parse(const GOptionEntry * optentry);
+
 #endif
--- ./wscript.orig	2010-06-19 20:25:14.000000000 +0300
+++ ./wscript	2010-06-25 19:24:13.000000000 +0300
@@ -110,7 +110,7 @@
 	'src/highlighting.c', 'src/interface.c', 'src/keybindings.c',
 	'src/keyfile.c', 'src/log.c', 'src/main.c', 'src/msgwindow.c', 'src/navqueue.c', 'src/notebook.c',
 	'src/plugins.c', 'src/pluginutils.c', 'src/prefix.c', 'src/prefs.c', 'src/printing.c', 'src/project.c',
-	'src/sciwrappers.c', 'src/search.c', 'src/socket.c', 'src/stash.c',
+	'src/sciwrappers.c', 'src/search.c', 'src/sm.c', 'src/socket.c', 'src/stash.c',
 	'src/symbols.c',
 	'src/templates.c', 'src/toolbar.c', 'src/tools.c', 'src/sidebar.c',
 	'src/ui_utils.c', 'src/utils.c' ]
@@ -222,6 +222,12 @@
 		gtk_version = 'Unknown'
 	conf.check_cfg(package='gio-2.0', uselib_store='GIO', args='--cflags --libs', mandatory=False)
 
+	# Find libSM
+	if not Options.options.no_libsm:
+		sm_version = conf.check_cfg(package='sm', uselib_store='SM', args='--cflags --libs', mandatory=False)
+		if sm_version is None:
+			Options.options.no_libsm = True
+
 	# Windows specials
 	if is_win32:
 		if conf.env['PREFIX'] == tempfile.gettempdir():
@@ -270,6 +276,7 @@
 	conf_define_from_opt('HAVE_PLUGINS', not Options.options.no_plugins, None, 0)
 	conf_define_from_opt('HAVE_SOCKET', not Options.options.no_socket, None, 0)
 	conf_define_from_opt('HAVE_VTE', not Options.options.no_vte, None, 0)
+	conf_define_from_opt('HAVE_LIBSM', not Options.options.no_libsm, None, 0)
 
 	conf.write_config_header('config.h')
 
@@ -278,6 +285,7 @@
 	print_message(conf, 'Using GTK version', gtk_version)
 	print_message(conf, 'Build with GTK printing support', have_gtk_210 and 'yes' or 'no')
 	print_message(conf, 'Build with plugin support', Options.options.no_plugins and 'no' or 'yes')
+	print_message(conf, 'Build with X session management support', Options.options.no_libsm and 'no' or 'yes')
 	print_message(conf, 'Use virtual terminal support', Options.options.no_vte and 'no' or 'yes')
 	if svn_rev != '-1':
 		print_message(conf, 'Compiling Subversion revision', svn_rev)
@@ -307,6 +315,9 @@
 		opt.add_option('--disable-vte', action='store_true', default=target_is_win32(os.environ),
 			help='compile without support for an embedded virtual terminal [[default: No]',
 			dest='no_vte')
+		opt.add_option('--disable-libsm', action='store_true', default=target_is_win32(os.environ),
+			help='compile without X session management support [[default: No]',
+			dest='no_libsm')
 		opt.add_option('--enable-gnu-regex', action='store_true', default=False,
 			help='compile with included GNU regex library [default: No]', dest='gnu_regex')
 		# Paths
@@ -385,7 +396,7 @@
 		source			= geany_sources,
 		includes		= '. src/ scintilla/include/ tagmanager/include/',
 		defines			= 'G_LOG_DOMAIN="Geany"',
-		uselib			= 'GTK GIO WIN32',
+		uselib			= 'GTK GIO SM WIN32',
 		uselib_local	= 'scintilla tagmanager',
 		add_objects		= 'geany-rc' if is_win32 else None
 	)
